import json
from pathlib import Path
from typing import Tuple

from filelock import FileLock
from nltk.tree import Tree
from ruamel.yaml import YAML
from stanfordcorenlp import StanfordCoreNLP

BELLA_CONFIG_FP = Path.home().joinpath('.Bella', 'config.yaml')
BELLA_CONFIG_LOCK_FP = Path.home().joinpath('.Bella', 'config.yaml.lock')

class StanfordNlp(object):
    '''
    Singleton Class instance
    '''

    instance = None

    @staticmethod
    def get_config() -> Tuple[str, int]:
        hostname = 'http://localhost'
        port = 9000
        yaml = YAML()
        config_data = {}
        BELLA_CONFIG_FP.parent.mkdir(parents=True, exist_ok=True)

        lock = FileLock(BELLA_CONFIG_LOCK_FP)
        with lock.acquire(60):
            if BELLA_CONFIG_FP.exists():
                with BELLA_CONFIG_FP.open('r') as config_file:
                    config_data = yaml.load(config_file)
                    if 'stanford_core_nlp' in config_data:
                        stanford_config = config_data['stanford_core_nlp']
                        if 'hostname' in stanford_config:
                            hostname = stanford_config['hostname']
                        if 'port' in stanford_config:
                            port = stanford_config['port']
            
            config_data['stanford_core_nlp'] = {}
            config_data['stanford_core_nlp']['hostname'] = hostname
            config_data['stanford_core_nlp']['port'] = port
            with BELLA_CONFIG_FP.open('w') as config_file:
                yaml.dump(config_data, config_file)
        return hostname, port

    def __new__(cls):
        if StanfordNlp.instance is None:
            hostname, port = cls.get_config()
            StanfordNlp.instance = StanfordCoreNLP(hostname, port)
        return StanfordNlp.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)


def tokenise(text):

    stanford_nlp = StanfordNlp()
    output_dict = stanford_nlp.annotate(text, {'annotators': 'ssplit,tokenize',
                                               'tokenize.language': 'English',
                                               'outputFormat': 'json'})
    output_dict = json.loads(output_dict, strict=False)
    tokens = [token['originalText'] for s in output_dict['sentences']
              for token in s['tokens']]
    return tokens


def constituency_parse(text):
    '''
    :param text: The text you want to parse
    :type text: String
    :returns: A list of parse trees where each tree is represented by \
    nltk.tree.Tree. Each parse tree is associated to a sentence in the text. \
    If one sentence in the text then the list will be of length 1.
    :rtype: list
    '''

    if text.strip() == '':
        raise ValueError('There has to be some text to parse. Text given {}'
                         .format(text))
    stanford_nlp = StanfordNlp()
    output_dict = stanford_nlp.annotate(text, {'annotators': 'pos,parse',
                                               'tokenize.language': 'English',
                                               'outputFormat': 'json'})
    output_dict = json.loads(output_dict)
    parse_trees = [Tree.fromstring(sent['parse'])
                   for sent in output_dict['sentences']]
    return parse_trees


def dependency_parse(text, dep_type='basicDependencies'):
    '''
    :param text: The text you want to parse
    :param dep_type: The dependency type to use either: 'basicDependencies', \
    'enhancedDependencies', or 'enhancedPlusPlusDependencies'. See for more \
    details \
    https://nlp.stanford.edu/~sebschu/pubs/schuster-manning-lrec2016.pdf
    :type text: String
    :type dep_type: String. Default basicDependencies
    '''

    if text.strip() == '':
        raise ValueError('There has to be some text to parse. Text given {}'
                         .format(text))

    stanford_nlp = StanfordNlp()
    output_dict = stanford_nlp.annotate(text, {'annotators': 'pos,depparse',
                                               'tokenize.language': 'English',
                                               'outputFormat': 'json'})
    sentences = json.loads(output_dict)['sentences']
    tokens_dicts = []
    dep_dicts = []
    for sentence in sentences:
        tokens_dict = {token_data['index']: token_data
                       for token_data in sentence['tokens']}
        dep_dict = {}
        for dep_data in sentence[dep_type]:
            dep_rel = dep_data['dep']
            dep_token_index = dep_data['governor']
            current_token_index = dep_data['dependent']
            dep_dict[current_token_index] = (dep_rel, dep_token_index)
        tokens_dicts.append(tokens_dict)
        dep_dicts.append(dep_dict)
    return dep_dicts, tokens_dicts



'''
A set of functions which either produce contexts and related targets based on
syntactic parsing. Or functions that create
'''
import re

from bella.data_types import Target
from bella.dependency_parsers import stanford


def normalise_target(target, text, sorted_target_spans, renormalise=False,
                     parser=None):

    def add_target_to_text(target):
        target_added_text = text
        for start, end in sorted_target_spans:
            start_text = target_added_text[: start]
            end_text = target_added_text[end :]
            start_text += ' {} '.format(target)
            target_added_text = start_text + end_text
        return ' '.join(target_added_text.split())

    def replace_target_in_text(target, new_target):
        return text.replace(target, new_target)

    def check_target_unique(target, text, num_target_spans):
        escaped_target = re.escape(target)
        num_target_occurences = len(re.findall(escaped_target, text))
        if num_target_occurences != num_target_spans:
            return False
        return True

    norm_target = target.strip()
    norm_text = text
    num_target_spans = len(sorted_target_spans)

    split_target = target.split()
    num_spaces_in_target = len(split_target)
    # Converts words such LG Flat Screen into LG_FlatScreen This is done as the
    # parser won't keep a word as a word if it is LG_Flat_Screen
    if renormalise:
        norm_target = ''.join(split_target)
        if norm_target[0] == '#' or norm_target[0] == '@':
            if '#' in norm_target[1:]:
                norm_target = norm_target.replace('#', '')
                norm_target = '#{}'.format(norm_target)
            elif '@' in norm_target[1:]:
                norm_target = norm_target.replace('@', '')
                norm_target = '@{}'.format(norm_target)

            if parser == stanford:
                norm_target = norm_target.replace('@', '')
        if norm_target[-1] == '#':
            norm_target = norm_target.replace('#', '')

    elif num_spaces_in_target > 2:
        first_word = split_target[0]
        joined_rest_words = ''.join(split_target[1:])
        norm_target = '_'.join([first_word, joined_rest_words])
    # Keeps the word as a whole word.
    elif num_spaces_in_target == 2:
        temp_split_target = []
        for word_index, split_word in enumerate(split_target):
            if word_index != 0:
                split_word = split_word.replace('#', '')
            temp_split_target.append(split_word)
        norm_target = '_'.join(temp_split_target)
    # Gets rid of anything this is not a word @ or #
    norm_target = re.sub(r'[^\w@#]', '', norm_target)

    # Put the normalised targets into the text
    norm_text = add_target_to_text(norm_target)

    #if parser == stanford:
    #  norm_target = norm_target.replace('@', '')

    # Checks a word that is the normalised target word exists in the text and
    # if so changes it to a word that is not by putting dollar signs around it.
    if not check_target_unique(norm_target, norm_text, num_target_spans):
        if parser == stanford:
            norm_target = norm_target.replace('@', '')
            norm_target = norm_target.replace('#', '')
        elif '@' in norm_target:
            norm_target = norm_target.replace('@', '')
        else:
            if renormalise and len(split_target) == 1:
                if norm_target[0] == '#':
                    norm_target = '${}$'.format(norm_target[1:])
            else:
                norm_target = '${}$'.format(norm_target)


        norm_text = add_target_to_text(norm_target)
        if not check_target_unique(norm_target, norm_text, num_target_spans):
            if parser == stanford:
                norm_target = '<{}>'.format(norm_target)
                norm_text = add_target_to_text(norm_target)
                if check_target_unique(norm_target,
                                       norm_text, num_target_spans):
                    return norm_target
            elif '$' not in norm_target:
                norm_target = '${}$'.format(norm_target)
                norm_text = add_target_to_text(norm_target)
                if check_target_unique(norm_target,
                                       norm_text, num_target_spans):
                    return norm_target
            raise Exception('Normalised word {} occurs in the text more times '\
                            'than it spans {}. Text {}'\
                            .format(norm_target, num_target_spans, norm_text))
    elif parser == stanford:
        if '@' in norm_target:
            norm_target

    return norm_target

def target_normalisation(target_dict, renormalise=False, parser=None):
    '''
    Given a target instance it normalises the target by removing whitespaces
    between target words and inserting `_`. Then inserting the normalised word
    into where the target spans appear and adding whitespace around the target
    word incase other words are joined on. Returns the text with the inserted
    normalised words and the normalised target word.

    :param target_dict: target instance.
    :type target_dict: Target
    :returns: Tuple of two Strings containing. Text with normalised targets and \
    the normalised target
    :rtype: tuple
    '''

    sorted_spans = sorted(target_dict['spans'], key=lambda span: span[0],
                          reverse=True)
    org_text = target_dict['text']
    target = target_dict['target']


    target = normalise_target(target, org_text, sorted_spans, renormalise,
                              parser=parser)
    for start_index, end_index in sorted_spans:
        start_text = org_text[: start_index]
        end_text = org_text[end_index :]
        start_text += ' {} '.format(target)
        org_text = start_text + end_text
    org_text = ' '.join(org_text.split())

    return org_text, target

def normalise_context(target_dicts, lower, renormalise=False, parser=None):
    '''
    Given a list of target dicts and if the text should be lower cased returns
    all of the text and targets within those target dicts as lists where the
    text and targets have been normalised to ensure the targets within the
    text can be identified.

    :param target_dicts: list of dicts
    :param lower: state if the text within the dicts should be lower cased
    :type target_dicts: list
    :type lower: bool
    :returns: A tuple of length two which contain a list of normalised texts and \
    targets.
    :rtype: list
    '''

    # Normalise the target and text
    all_text = []
    all_norm_targets = []
    for target_dict in target_dicts:
        norm_text, norm_target = target_normalisation(target_dict, renormalise,
                                                      parser=parser)
        if lower:
            norm_text = norm_text.lower()
            norm_target = norm_target.lower()
        all_text.append(norm_text)
        all_norm_targets.append(norm_target)
    return all_text, all_norm_targets

def dependency_relation_context(target_dicts, parser, lower=False,
                                n_relations=(1, 1)):
    '''
    Given a list of target dicts where each target dict has a sentence that
    contains one or more of the same target. Returns a list of a list of Strings
    where each String is associated to a target within the target sentence.
    The String is a concatenation of n_relations depth of dependency relations
    where each relation is a child of the target. e.g. n_relations = (1, 1)
    will return a String of the concatenation of the children of the target
    within the dependency tree. n_relations = (1, 2) will return the children
    of the target and the children of the children.

    :param target_dicts: list of dictionaries where each dictionary is associated \
    to a target sentence.
    :param parser: function that performs dependency parsing
    :param lower: Whether to lower case the text
    :param n_relations: The depth of the dependency relation text from the target \
    to return. Represented as a tuple of two ints the first defining the \
    starting depth the second end depth e.g. (1, 2) will return depths one and \
    two of dependency tree.
    :type target_dicts: list
    :type parser: String
    :type lower: bool. Default False.
    :type n_relations: tuple. Default (1, 1).
    :returns: A list of a list of Strings where each String represents a specific \
    target word within a target sentence dependency related text at n_relations \
    depth.
    :rtype: list
    '''

    # Normalise the target and text
    targets = [target_dict['target'] for target_dict in target_dicts]
    norm_texts, norm_targets = normalise_context(target_dicts, lower,
                                                 parser=parser)
    # Get contexts
    all_dependency_tokens = parser(norm_texts)
    all_contexts = []

    for index, dependency_tokens in enumerate(all_dependency_tokens):
        for attempts in range(1, 3):
            contexts = []
            norm_target = norm_targets[index]
            # This only happens if the first normalization does not work
            if attempts == 2:
                text, norm_target = normalise_context([target_dicts[index]],
                                                      lower=lower, parser=parser,
                                                      renormalise=True)
                norm_target = norm_target[0]
                dependency_tokens = parser(text)[0]
            for dependency_token in dependency_tokens:
                current_token = targets[index]
                if lower:
                    current_token = current_token.lower()
                if dependency_token.token == norm_target:
                    all_related_words = dependency_token.get_n_relations(n_relations)
                    related_text = ' '.join(all_related_words)
                    related_text = related_text.replace(current_token, norm_target)
                    contexts.append(related_text)

            rel_target = target_dicts[index]
            valid_num_targets = len(rel_target['spans'])
            if valid_num_targets != len(contexts):
                if attempts == 1:
                    continue
                raise ValueError('The number of identified targets `{}` not equal '\
                                 'to the number of targets in the data `{}`'\
                                 .format(contexts, rel_target))
            # Ensure the returned data type is consistent
            if contexts == []:
                raise ValueError('This should not happen as each data type should '\
                                 'have a target {}'.format(rel_target))
            all_contexts.append(contexts)
            break
    return all_contexts

def dependency_context(target_dicts, parser, lower=False):
    '''
    Given a list of target dicts it will normalise the target word to ensure
    that it is seperated and if it is a multi word target join the target words
    together to ensure when it is processed by the dependency parser it is treated
    as a singular word.

    Given a list of target sentences returns a list of contexts where each contexts
    is associated to a target sentence and each contexts contains a target context
    for each target word in the target sentence. A target context is a dict which
    contains `text` and `span` keys where the values correspond to all the
    dependency related words as a String and the span are the indexs to the target
    word within the text.

    :param target_dicts: list of dictionaries where each dictionary is associated \
    to a target sentence.
    :param parser: function that performs dependency parsing
    :param lower: Whether to lower case the texts before processing them with the \
    parser.
    :type target_dicts: list
    :type parser: function
    :type lower: bool Default False
    :returns: A list of a list of dicts where each list is contains many contexts.
    :rtype: list
    '''

    # Normalise the target and text
    targets = [target_dict['target'] for target_dict in target_dicts]
    norm_texts, norm_targets = normalise_context(target_dicts, lower,
                                                 parser=parser)
    # Get contexts
    all_dependency_tokens = parser(norm_texts)
    all_contexts = []
    for index, dependency_tokens in enumerate(all_dependency_tokens):
        for attempts in range(1, 3):
            contexts = []
            norm_target = norm_targets[index]
            # This only happens if the first normalization does not work
            if attempts == 2:
                text, norm_target = normalise_context([target_dicts[index]],
                                                      lower=lower, parser=parser,
                                                      renormalise=True)
                norm_target = norm_target[0]
                dependency_tokens = parser(text)[0]

            for dependency_token in dependency_tokens:
                current_target = targets[index]
                if lower:
                    current_target = current_target.lower()

                if dependency_token.token == norm_target:
                    norm_org = (norm_target, current_target)
                    text_span = dependency_token\
                                .connected_target_span(renormalise=norm_org)
                    connected_text, target_span = text_span
                    contexts.append({'text' : connected_text,
                                     'span' : target_span})
            rel_target = target_dicts[index]
            valid_num_targets = len(rel_target['spans'])
            if valid_num_targets != len(contexts):
                if attempts == 1:
                    continue
                raise ValueError('The number of identified targets `{}` not equal '\
                                 'to the number of targets in the data `{}`'\
                                 ' norm target {}'\
                                 .format(contexts, rel_target, norm_target))
            # Ensure the returned data type is consistent
            if contexts == []:
                raise ValueError('This should not happen as each data type should '\
                                 'have a target {}'.format(rel_target))
            all_contexts.append(contexts)
            break
    return all_contexts

def context(all_context_dicts, specific_context, inc_target=False):
    '''
    Returns a list of a list of Strings based on the location of the target word
    in the text within the target dict (NOTE the target word can occur more than
    once hence why a list is returned as the context is returned for each
    occurence). Context can be one of the following:

    1. left - left of the target occurence.
    2. right - right of the target occurence.
    3. target - target word/words of each target occurence.
    4. full - whole text repeated for each occurence.

    If the target only occur once in the text then for that text occurence the
    length of the list will be one.

    :param all_context_dicts: A list of dicts where each dict contains text and \
    span keys.
    :param specific_context: String specifying the context e.g. left.
    :param inc_target: Whether to include the target word in the context text. \
    (Only applies for left and right context.)
    :type all_context_dicts: list
    :type specific_context: String
    :type inc_target: Boolean Default False
    :returns: A list of of a list of context strings
    :rtype: list
    '''

    acceptable_contexts = {'left', 'right', 'target', 'full'}
    if specific_context not in acceptable_contexts:
        raise ValueError('context parameter can only be one of the following {}'\
                         ' not {}'.format(acceptable_contexts, context))
    all_contexts = []
    for context_dicts in all_context_dicts:
        contexts = []
        for context_dict in context_dicts:
            text = context_dict['text']
            target_span = context_dict['span']
            start_char = target_span[0]
            end_char = target_span[1]
            if specific_context == 'left':
                if inc_target:
                    contexts.append(text[:end_char])
                else:
                    contexts.append(text[:start_char])
            elif specific_context == 'right':
                if inc_target:
                    contexts.append(text[start_char:])
                else:
                    contexts.append(text[end_char:])
            elif specific_context == 'target':
                contexts.append(text[start_char:end_char])
            elif specific_context == 'full':
                contexts.append(text)
            else:
                raise ValueError('context parameter should only be `right` or '\
                                 '`left` not {} there must be a logic error'\
                                 .format(specific_context))
        all_contexts.append(contexts)
    return all_contexts



'''
Contains classes that train and/or load semantic vectors. All classes are sub
classes of WordVectors

Classes:

1. WordVectors - Base class of all classes within this module. Ensures
consistent API for all word vectors classes.
2. GensimVectors - Creates `Word2Vec <https://arxiv.org/pdf/1301.3781.pdf>`_
and `FastText <https://arxiv.org/abs/1607.04606>`_ vectors.
3. PreTrained - Creates a Wordembedding for those that are stored in TSV files
where the first item in the line is the word and the rest of the tab sep values
are its vector representation. Currently loads the Tang et al. vectors
`from <https://github.com/bluemonk482/tdparse/tree/master/resources/wordemb/sswe>`_
'''

from collections import defaultdict
import os
import math
import types
import tempfile
import zipfile
from pathlib import Path

import numpy as np
import requests
from gensim.models.keyedvectors import KeyedVectors
from gensim.models import word2vec
from gensim.models.wrappers import FastText
from gensim.scripts.glove2word2vec import glove2word2vec
from tqdm import tqdm

BELLA_VEC_DIR = Path.home().joinpath('.Bella', 'Vectors')


class WordVectors(object):
    '''
    Base class for all WordVector classes. Contains the following instance
    attributes:

    1. vector_size - Size of the word vectors e.g. 100
    2. index2word - Mapping between index number and associated word
    3. index2vector - mapping between index and vector
    4. word2index - Mapping between word and associated index
    5. name - This is used to identify the model when reading cross validation \
    results from models, used mainly for debugging. Default is None. Used
    6. unknown_vector - The vector that is returned for any unknwon words and \
    for the index=0.
    7. unknown_word - The word that is returned for the 0 index. Default is \
    `<unk>`
    8. unit_length - If the vectors when returned are their unit norm value \
    instead of their raw values.
    9. unknown_index - The index of the unknown word normally 0
    10. padding_word - The word (<pad>) that defines padding indexs.
    11. padding_index - index of the padding word
    12. padding vector - padding vector for the padding word.

    padding index, word and vector are equal to the unknown equilavents if
    padding_value = None in the constructor. Else padding index = 0, word = <pad>
    and vector is what you have defined, this then moves the unknown index to
    vocab size + 1 and the everything else is the same. The idea behind the 2
    are that pad is used for padding and unknown is used for words that are
    truely unknown therefore allowing you to only skip the pad vectors when
    training a model by using a masking function in keras.

    The index 0 is used as a special index to map to unknown words. Therefore \
    the size of the vocabularly is len(index2word) - 1.

    Following methods:

    1. :py:func:`bella.word_vectors.WordVectors.lookup_vector`
    '''
    def __init__(self, word2vector, name=None, unit_length=False,
                 padding_value=None, filter_words=None):
        self.filter_words = [] if filter_words is None else filter_words
        size_vector_list = self._check_vector_size(word2vector)
        self.vector_size, self._word2vector, self._word_list = size_vector_list
        self.name = '{}'.format(name)
        self.unit_length = unit_length
        # Method attributes
        self.unknown_word = self._unknown_word()
        self.unknown_vector = self._unknown_vector()
        self.unknown_index = 0
        self.padding_word = self.unknown_word
        self.padding_vector = None
        if padding_value is not None:
            self.unknown_index = len(self._word_list) + 1
            self.padding_vector = np.asarray([padding_value] *
                                             self.vector_size)
            self.padding_word = '<pad>'
        else:
            self.padding_vector = self.unknown_vector
        if self.padding_vector is None:
            raise ValueError('Padding Vector is None')
        self.index2word = self._index2word()
        self.word2index = self._word2index()
        self.index2vector = self._index2vector()
        self.embedding_matrix = self._embedding_matrix()

    def _keyed_vec_2_dict(self, key_vector):
        word_2_vec_dict = {}
        for word in key_vector.vocab:
            word_2_vec_dict[word] = key_vector[word]
        return word_2_vec_dict

    def _check_vector_size(self, word2vector):
        '''
        This finds the most common vector size in the word 2 vectors dictionary
        mapping. Normally they should all have the same mapping but it has been
        found that their could be some mistakes in pre-compiled word vectors
        therefore this function removes all words and vectors that do not conform
        to the majority vector size. An example of this would be if all the
        word to vector mappings are on dimension 50 but one word that one word
        would be removed from the dictionary mapping and not included in the
        word list returned.

        :param word2vector: A dictionary containing words as keys and their \
        associated vector representation as values.
        :type word2vector: dict or gensim.models.keyedvectors.KeyedVectors
        :returns: A tuple of length 3 containg 1. The dimension of the vectors, \
        2. The dictionary of word to vectors, 3. The list of words in the dictionary
        :rtype: tuple
        '''
        # Gensim does not used a dictionary but a special class
        if isinstance(word2vector, KeyedVectors):
            word2vector = self._keyed_vec_2_dict(word2vector)
            #accepted_words = word2vector.index2word
            #most_likely_size = word2vector[accepted_words[0]].shape[0]
            #return most_likely_size, word2vector, accepted_words
        vector_sizes = {}
        for _, vector in word2vector.items():
            vector_size = vector.shape[0]
            vector_sizes[vector_size] = vector_sizes.get(vector_size, 0) + 1
        most_likely_size = sorted(vector_sizes.items(), reverse=True,
                                  key=lambda size_freq: size_freq[0])[0][0]
        words_to_remove = []
        unk_word = self._unknown_word()
        accepted_words = []
        for word, vector in word2vector.items():
            if vector.shape[0] != most_likely_size:
                words_to_remove.append(word)
            elif self.filter_words != []:
                if word not in self.filter_words and word != unk_word:
                    words_to_remove.append(word)
                else:
                    accepted_words.append(word)
            else:
                accepted_words.append(word)
        for word in words_to_remove:
            del word2vector[word]
        return most_likely_size, word2vector, accepted_words

    @staticmethod
    def glove_txt_binary(glove_file_path: Path):
        '''
        Converts the Glove word embedding file which is a text file to a
        binary file that can be loaded through gensims
        KeyedVectors.load_word2vec_format method and deletes the text file
        version and returns the file path to the new binary file.

        :param glove_file_path: File path to the downloaded glove vector text \
        file.
        :type glove_file_path: String
        :returns: The file path to the binary file version of the glove vector
        :rtype: String
        '''
        # File path to the binary file
        binary_file_path = os.path.splitext(glove_file_path.name)
        binary_file_path = binary_file_path[0]
        binary_file_path += '.binary'
        binary_file_path = glove_file_path.parent.joinpath(binary_file_path)
        if binary_file_path.is_file():
            return str(binary_file_path.resolve())
        with tempfile.NamedTemporaryFile('w', encoding='utf-8') as temp_file:
            # Converts to word2vec file format
            print('Converting word vectors file from text to binary for '
                  'quicker load time')
            glove2word2vec(str(glove_file_path.resolve()), temp_file.name)
            word_vectors = KeyedVectors.load_word2vec_format(temp_file.name,
                                                             binary=False)
            # Creates the binary file version of the word vectors
            binary_file_path = str(binary_file_path.resolve())
            word_vectors.save_word2vec_format(binary_file_path, binary=True)
        # Delete the text version of the glove vectors
        glove_file_path.unlink()
        return binary_file_path

    def lookup_vector(self, word):
        '''
        Given a word returns the vector representation of that word. If the model
        does not have a representation for that word it returns word vectors
        unknown word vector (most models this is zeors)

        :param word: A word
        :type word: String
        :returns: The word vector for that word. If no word vector can be found
        returns a vector of zeros.
        :rtype: numpy.ndarray
        '''

        if isinstance(word, str):
            word_index = self.word2index[word]
            return self.index2vector[word_index]
        raise ValueError('The word parameter must be of type str not '
                         f'{type(word)}')

    def _index2word(self):
        '''
        The index starts at one as zero is a special value assigned to words that
        are padded. The word return for zero is defined by `padded_word` attribute.
        If the padded value is different to the unknown word value then the index
        will contain an extra index for the unknown word index which is the
        vocab size + 1 index.

        :returns: A dictionary matching word indexs to there corresponding words.
        Inverse of :py:func:`bella.word_vectors.GensimVectors.word2index`
        :rtype: dict
        '''

        index_word = {}
        index_word[0] = self.padding_word
        index = 1
        for word in self._word_list:
            if word == self.unknown_word:
                continue
            index_word[index] = word
            index += 1
        # Required as the unknown word might have been learned and in
        # self._word_list
        if self.unknown_index != 0:
            self.unknown_index = len(index_word)
        index_word[self.unknown_index] = self.unknown_word
        return index_word

    def _return_unknown_index(self):
        '''
        :returns: zero. Used as lambda is not pickleable
        :rtype: int
        '''
        return self.unknown_index

    def _word2index(self):
        '''
        If you have specified a special padded index vector then the <pad> word
        would match to index 0 and the vocab + 1 index will be <unk> else if
        no special pad index vector then vocab + 1 won't exist and <unk> will be
        0

        :returns: A dictionary matching words to there corresponding index.
        Inverse of :py:func:`bella.word_vectors.GensimVectors.index2word`
        :rtype: dict
        '''

        # Cannot use lambda function as it cannot be pickled
        word2index_dict = defaultdict(self._return_unknown_index)
        for index, word in self.index2word.items():
            word2index_dict[word] = index
        return word2index_dict

    def _index2vector(self):
        '''
        NOTE: the zero index is mapped to the unknown index unless padded vector
        is specified then zero index is padded index and vocab + 1 index is
        unknown index

        :returns: A dictionary of word index to corresponding word vector. Same
        as :py:func:`bella.word_vectors.GensimVectors.lookup_vector` but
        instead of words that are looked up it is the words index.
        :rtype: dict
        '''

        def unit_norm(vector):
            '''
            :param vector: A 1-dimension vector
            :type vector: numpy.ndarray
            :returns: The vector normalised to it's unit vector.
            :rtype: numpy.ndarray
            '''

            # Check it is not a zero vector
            if np.array_equal(np.zeros(self.vector_size), vector):
                return vector
            l2_norm = np.linalg.norm(vector)
            return vector / l2_norm

        index_vector = {}
        if self.unit_length:
            index_vector[0] = unit_norm(self.padding_vector)
            index_vector[self.unknown_index] = unit_norm(self.unknown_vector)
        else:
            index_vector[0] = self.padding_vector
            index_vector[self.unknown_index] = self.unknown_vector
        for index, word in self.index2word.items():
            if index == 0 or index == self.unknown_index:
                continue
            if self.unit_length:
                index_vector[index] = unit_norm(self._word2vector[word])
            else:
                index_vector[index] = self._word2vector[word]
        return index_vector

    def __repr__(self):
        return self.name

    def _unknown_vector(self):
        '''
        This is to be Overridden by sub classes if they want to return a custom
        unknown vector.

        :returns: A vector for all unknown words. In this case it is a zero
        vector.
        :rtype: numpy.ndarray
        '''

        return np.zeros(self.vector_size)

    def _unknown_word(self):
        '''
        :returns: The word that is returned for the 0 index.
        :rtype: String
        '''

        return '<unk>'

    def _embedding_matrix(self):
        '''
        The embedding matrix that can be used in Keras embedding layer as the
        weights. It is very much a simple lookup where the key is the word index.

        :retunrs: The embedding matrix of dimension (vocab_size + 1, vector_size) \
        where the vocab size is + 1 due to the unknown vector.
        :rtype: numpy.ndarray
        '''
        matrix = np.zeros((len(self.index2vector), self.vector_size),
                          dtype=np.float32)
        for index, vector in self.index2vector.items():
            try:
                matrix[index] = vector
            except Exception as e:
                word = self.index2word[index]
                print(f'{word} {index} {vector} {e}')
        return matrix


class GensimVectors(WordVectors):
    '''
    Class that can create one of the following word vector models:

    1. `Word2Vec <https://radimrehurek.com/gensim/models/word2vec.html>`_.
    2. `Fasttext <https://radimrehurek.com/gensim/models/wrappers/fasttext.html>`_

    private attributes:

    1. self._model = Gensim instance of the chosen model e.g. if word2vec
    was chosen then it would be `gensim.models.word2vec.Word2Vec`
    '''

    def __init__(self, file_path, train_data, name=None, model=None,
                 unit_length=False, padding_value=None,
                 filter_words=None, **kwargs):
        '''
        Trains or loads the model specified.

        :param file_path: Path to the saved model or Path to save the model
        once trained. Can be None if you only want to train.
        :param train_data: An object like a list that can be iterated that contains
        tokenised text, to train the model e.g. [['hello', 'how'], ['another']].
        Not required if `file_path` contains a trained model.
        :param model: The name of the model
        :param name: The name of given to the instance.
        :param unit_length: If the word vectors should be normalised to unit \
        vectors
        :param kwargs: The keyword arguments to give to the Gensim Model that is \
        being used i.e. keyword argument to `Word2Vec <https://radimrehurek.com/\
        gensim/models/word2vec.html#gensim.models.word2vec.Word2Vec>`_
        :type file_path: String
        :type train_data: iterable object e.g. list
        :type model: String
        :type name: String Default None
        :type unit_length: bool. Default False.
        :type kwargs: dict
        '''

        allowed_models = {'word2vec': word2vec.Word2Vec,
                          'fasttext': FastText}
        if model not in allowed_models:
            raise ValueError('model parameter has to be one of the following '
                             f'{allowed_models.keys()} not {model}')
        model = allowed_models[model]
        failed_to_load = True

        if isinstance(file_path, str):
            file_path = os.path.abspath(file_path)

        if isinstance(file_path, str):
            if os.path.isfile(file_path):
                try:
                    self._model = model.load(file_path)
                    failed_to_load = False
                except EOFError:
                    failed_to_load = True

        if hasattr(train_data, '__iter__') and failed_to_load:
            # Generators throws an error in Gensim
            if isinstance(train_data, types.GeneratorType):
                train_data = map(lambda x: x, train_data)
            self._model = model(train_data, **kwargs)
            if isinstance(file_path, str):
                self._model.save(file_path)
                print('{} model has been saved to {}'.format(model.__name__,
                                                             file_path))
        elif failed_to_load:
            raise Exception('Cannot create model as there is no path to extract '\
                            'a model from {} or any data to train on which has '\
                            'to have the __iter__ function {}'\
                            .format(file_path, train_data))
        super().__init__(self._model.wv, name=name, unit_length=unit_length,
                         padding_value=padding_value,
                         filter_words=filter_words)


class VoVectors(GensimVectors):

    def download(self, skip_conf):
        vo_folder = BELLA_VEC_DIR.joinpath('Vo Word Vectors')
        vo_folder.mkdir(parents=True, exist_ok=True)
        current_vector_files = set([vo_file.name for vo_file
                                    in vo_folder.iterdir()])
        all_vector_files = set(['c10_w3_s100',
                                'c10_w3_s100.syn0.npy',
                                'c10_w3_s100.syn1.npy'])
        interset = all_vector_files.intersection(current_vector_files)
        # If the files in the folder aren't all the glove files that would be
        # downloaded re-download the zip and unzip the files.
        if interset != all_vector_files:
            can_download = 'yes'
            if not skip_conf:
                download_msg = 'We are going to download the Vo Word vectors '\
                               'this is a download of 120MB '\
                               'Would you like to continue? If so type '\
                               '`yes`\n>> '
                can_download = input(download_msg)

            if can_download.strip().lower() == 'yes':
                base_url = 'https://github.com/bluemonk482/tdparse/raw/'\
                           'master/resources/wordemb/w2v/'
                link_locations = [(f'{base_url}c10_w3_s100',
                                   vo_folder.joinpath('c10_w3_s100')),
                                  (f'{base_url}c10_w3_s100.syn0.npy',
                                   Path(vo_folder, 'c10_w3_s100.syn0.npy')),
                                  (f'{base_url}c10_w3_s100.syn1.npy',
                                   Path(vo_folder, 'c10_w3_s100.syn1.npy'))]
                print('Downloading Vo vectors')
                for download_link, file_location in link_locations:
                    # Reference:
                    # http://docs.python-requests.org/en/master/user/quickstart/#raw-response-content
                    with file_location.open('wb') as vo_file:
                        request = requests.get(download_link, stream=True)
                        total_size = int(request.headers.get('content-length',
                                                             0))
                        for chunk in tqdm(request.iter_content(chunk_size=128),
                                          total=math.ceil(total_size//128)):
                            vo_file.write(chunk)
            else:
                raise Exception('Vo vectors not downloaded therefore '
                                'cannot load them')
        return str(vo_folder.joinpath('c10_w3_s100').resolve())

    def __init__(self, name=None, unit_length=False,
                 padding_value=None, skip_conf=False,
                 filter_words=None):
        vector_file = self.download(skip_conf)

        if name is None:
            name = 'w2v'
        super().__init__(vector_file, train_data=None, name=name,
                         model='word2vec', unit_length=unit_length,
                         padding_value=padding_value,
                         filter_words=filter_words)



class PreTrained(WordVectors):
    '''
    Class that loads word vectors that have been pre-trained.

    All pre-trained word vectors have to follow the following conditions:

    1. New word vector on each line
    2. Each line is tab seperated (by default but a tab is just delimenter \
    which can be changed by setting delimenter argument in the constructor)
    3. The first tab sperated value on the line is the word
    4. The rest of the tab seperated values on that line represent the values
    for the associtaed word.
    '''

    def __init__(self, file_path, name=None, unit_length=False,
                 delimenter='\t', padding_value=None, filter_words=None):
        '''
        :param file_path: The file path to load the word vectors from
        :param name: The name given to the instance.
        :param unit_length: If the word vectors should be normalised to unit \
        vectors
        :param delimenter: The value to be used to split the values in each line \
        of the word vectors.
        :type file_path: String
        :type name: String Default None
        :type unit_length: bool. Default False
        :type delimenter: String. Default `\t`
        '''
        if not isinstance(file_path, str):
            raise TypeError('The type of the file path should be str not {}'\
                            .format(type(file_path)))
        file_path = os.path.abspath(file_path)
        if not os.path.isfile(file_path):
            raise ValueError('There is no file at file path {}'.format(file_path))

        word2vector = {}
        with open(file_path, 'r') as data_file:
            for org_line in data_file:
                line = org_line.strip()
                word_values = line.split(delimenter)
                word = word_values[0]
                # This attempts to remove words that have whitespaces in them
                # a sample of this problem can be found within the Glove
                # Common Crawl 840B vectors where \xa0name@domain.com ==
                # name@domain.com after strip is applied and they have different
                # vectors
                if word in word2vector:
                    org_word = org_line.split(delimenter)[0]
                    if word != org_word:
                        continue
                    else:
                        del word2vector[word]
                word_vector = np.asarray(word_values[1:], dtype='float32')
                if word in word2vector:
                    dict_vector = word2vector[word]
                    raise KeyError('{} already has a vector in the word vector '\
                                   'dict. Vector in dict {} and alternative vector {}'\
                                   .format(word, dict_vector, word_vector))
                else:
                    word2vector[word] = word_vector
        super().__init__(word2vector, name=name, unit_length=unit_length,
                         padding_value=padding_value,
                         filter_words=filter_words)

    def _unknown_vector(self):
        '''
        Overrides. Instead of returnning zero vector it return the vector for
        the word `<unk>`.

        :returns: The vector for the word `<unk>`
        :rtype: numpy.ndarray
        '''

        return self._word2vector['<unk>']


class SSWE(PreTrained):

    def download(self, skip_conf):
        '''
        '''
        sswe_folder = BELLA_VEC_DIR.joinpath('SSWE')
        sswe_folder.mkdir(parents=True, exist_ok=True)
        sswe_fp = sswe_folder.joinpath('sswe')
        # If the files in the folder aren't all the SSWE files that would be
        # downloaded re-download the zip and unzip the files.
        if not sswe_fp.is_file():
            can_download = 'yes'
            if not skip_conf:
                download_msg = 'We are going to download the SSWE vectors '\
                               'this is a download of 74MB '\
                               'Would you like to continue? If so type '\
                               '`yes`\n>> '
                can_download = input(download_msg)

            if can_download.strip().lower() == 'yes':
                download_link = 'https://github.com/bluemonk482/tdparse/raw/'\
                                'master/resources/wordemb/sswe/sswe-u.txt'
                # Reference:
                # http://docs.python-requests.org/en/master/user/quickstart/#raw-response-content
                with sswe_fp.open('wb') as sswe_file:
                    request = requests.get(download_link, stream=True)
                    total_size = int(request.headers.get('content-length', 0))
                    print('Downloading SSWE vectors')
                    for chunk in tqdm(request.iter_content(chunk_size=128),
                                      total=math.ceil(total_size//128)):
                        sswe_file.write(chunk)
            else:
                raise Exception('SSWE vectors not downloaded therefore '
                                'cannot load them')
            sswe_folder_files = list(sswe_folder.iterdir())
            if not sswe_fp.is_file():
                raise Exception('Error in either downloading the SSWE vectors'
                                ' or file path names. Files in the SSWE '
                                f'folder {sswe_folder_files} and where the '
                                f'SSWE file should be {str(sswe_fp)}')
        return str(sswe_fp.resolve())

    def __init__(self, name=None, unit_length=False, skip_conf=False,
                 padding_value=None, filter_words=None):

        vector_file = self.download(skip_conf)

        if name is None:
            name = 'sswe'
        super().__init__(vector_file, name=name, unit_length=unit_length,
                         padding_value=padding_value,
                         filter_words=filter_words)



class GloveTwitterVectors(WordVectors):

    def download(self, skip_conf):
        '''
        This method checks if the
        `Glove Twitter word vectors \
        <https://nlp.stanford.edu/projects/glove/>`_
        are already in the repoistory if not it downloads and unzips the word
        vectors if permission is granted and converts them into a gensim
        KeyedVectors binary representation.

        :param skip_conf: Whether to skip the permission step as it requires \
        user input. True to skip permission.
        :type skip_conf: bool
        :returns: A dict containing word vector dimension as keys and the \
        absolute path to the vector file.
        :rtype: dict
        '''

        glove_folder = BELLA_VEC_DIR.joinpath('glove_twitter')
        glove_folder.mkdir(parents=True, exist_ok=True)
        current_glove_files = set([glove_file.name for glove_file
                                   in glove_folder.iterdir()])
        all_glove_files = set(['glove.twitter.27B.25d.binary',
                               'glove.twitter.27B.50d.binary',
                               'glove.twitter.27B.100d.binary',
                               'glove.twitter.27B.200d.binary'])
        interset = all_glove_files.intersection(current_glove_files)
        # If the files in the folder aren't all the glove files that would be
        # downloaded re-download the zip and unzip the files.
        if interset != all_glove_files:
            can_download = 'yes'
            if not skip_conf:
                download_msg = 'We are going to download the glove vectors '\
                               'this is a large download of 1.4GB and takes '\
                               '5.4GB of disk space after being unzipped. '\
                               'Would you like to continue? If so type '\
                               '`yes`\n>> '
                can_download = input(download_msg)

            if can_download.strip().lower() == 'yes':
                download_link = 'http://nlp.stanford.edu/data/glove.twitter.'\
                                '27B.zip'
                glove_zip_path = glove_folder.joinpath('glove_zip.zip')
                # Reference:
                # http://docs.python-requests.org/en/master/user/quickstart/#raw-response-content
                with glove_zip_path.open('wb') as glove_zip_file:
                    request = requests.get(download_link, stream=True)
                    total_size = int(request.headers.get('content-length', 0))
                    print('Downloading Glove Twitter vectors')
                    for chunk in tqdm(request.iter_content(chunk_size=128),
                                      total=math.ceil(total_size//128)):
                        glove_zip_file.write(chunk)
                print('Unzipping word vector download')
                glove_zip_path = str(glove_zip_path.resolve())
                with zipfile.ZipFile(glove_zip_path, 'r') as glove_zip_file:
                    glove_zip_file.extractall(path=glove_folder)
            else:
                raise Exception('Glove Twitter vectors not downloaded '
                                'therefore cannot load them')

        def add_full_path(file_name):
            file_path = glove_folder.joinpath(file_name)
            return self.glove_txt_binary(file_path)

        return {25: add_full_path('glove.twitter.27B.25d.txt'),
                50: add_full_path('glove.twitter.27B.50d.txt'),
                100: add_full_path('glove.twitter.27B.100d.txt'),
                200: add_full_path('glove.twitter.27B.200d.txt')}

    def __init__(self, dimension, name=None, unit_length=False,
                 skip_conf=False, padding_value=None,
                 filter_words=None):
        '''
        :param dimension: Dimension size of the word vectors you would like to \
        use. Choice: 25, 50, 100, 200
        :param skip_conf: Whether to skip the permission step for downloading \
        the word vectors as it requires user input. True to skip permission.
        :type dimension: int
        :type skip_conf: bool. Default False
        '''

        dimension_file = self.download(skip_conf)
        if not isinstance(dimension, int):
            raise TypeError('Type of dimension has to be int not {}'
                            .format(type(dimension)))
        if dimension not in dimension_file:
            raise ValueError('Dimension avliable are the following {}'
                             .format(list(dimension_file.keys())))
        if name is None:
            name = f'glove twitter {dimension}d'
        vector_file = dimension_file[dimension]
        print(f'Loading {name} from file')
        glove_key_vectors = KeyedVectors.load_word2vec_format(vector_file,
                                                              binary=True)
        super().__init__(glove_key_vectors, name=name, unit_length=unit_length,
                         padding_value=padding_value,
                         filter_words=filter_words)

    def _unknown_vector(self):
        '''
        This is to be Overridden by sub classes if they want to return a custom
        unknown vector.

        :returns: A vector for all unknown words. In this case it is a zero
        vector.
        :rtype: numpy.ndarray
        '''

        return np.zeros(self.vector_size, dtype=np.float32)


class GloveCommonCrawl(WordVectors):

    def download(self, skip_conf, version):
        '''
        This method checks if either the `Glove Common Crawl \
        <https://nlp.stanford.edu/projects/glove/>`_ 840 or 42 Billion token
        word vectors were downloaded already into the repoistory if not it
        downloads and unzips the 300 Dimension word vector if permission is
        granted.

        :param skip_conf: Whether to skip the permission step as it requires \
        user input. True to skip permission.
        :param version: Choice of either the 42 or 840 Billion token 300 \
        dimension common crawl glove vectors. The values can be only 42 or \
        840 and default is 42.
        :type skip_conf: bool
        :type version: int
        :returns: The filepath to the 300 dimension word vector
        :rtype: String
        '''

        glove_folder = BELLA_VEC_DIR.joinpath(f'glove_common_crawl_{version}b')
        glove_folder.mkdir(parents=True, exist_ok=True)
        glove_file_name = f'glove.{version}B.300d'
        glove_txt_fp = glove_folder.joinpath(glove_file_name + '.txt')
        glove_binary_fp = glove_folder.joinpath(glove_file_name + '.binary')
        # If the files in the folder aren't all the glove files that would be
        # downloaded re-download the zip and unzip the files.
        if not glove_binary_fp.is_file() and not glove_txt_fp.is_file():
            can_download = 'yes'
            if not skip_conf:
                download_msg = 'We are going to download the glove vectors '\
                               'this is a large download of ~2GB and takes '\
                               '~5.6GB of diskspace after being unzipped. '\
                               'Would you like to continue? If so type '\
                               '`yes`\n>> '
                can_download = input(download_msg)

            if can_download.strip().lower() == 'yes':
                zip_file_name = f'glove.{version}B.300d.zip'
                download_link = 'http://nlp.stanford.edu/data/' + zip_file_name

                glove_zip_path = glove_folder.joinpath(zip_file_name)

                # Reference:
                # http://docs.python-requests.org/en/master/user/quickstart/#raw-response-content
                with glove_zip_path.open('wb') as glove_zip_file:
                    request = requests.get(download_link, stream=True)
                    total_size = int(request.headers.get('content-length', 0))
                    print(f'Downloading Glove {version}B vectors')
                    for chunk in tqdm(request.iter_content(chunk_size=128),
                                      total=math.ceil(total_size//128)):
                        glove_zip_file.write(chunk)
                print('Unzipping word vector download')
                glove_zip_path = str(glove_zip_path.resolve())
                with zipfile.ZipFile(glove_zip_path, 'r') as glove_zip_file:
                    glove_zip_file.extractall(path=glove_folder)
            else:
                raise Exception(f'Glove Common Crawl {version}b vectors '
                                'not downloaded therefore cannot load them')
            glove_folder_files = list(glove_folder.iterdir())
            if not glove_txt_fp.is_file():
                raise Exception('Error in either downloading the glove vectors'
                                ' or file path names. Files in the glove '
                                f'folder {glove_folder_files} and where the '
                                f'golve file should be {str(glove_txt_fp)}')
        return self.glove_txt_binary(glove_txt_fp)

    def __init__(self, version=42, name=None, unit_length=False,
                 skip_conf=False, padding_value=None,
                 filter_words=None):
        '''
        :param version: Choice of either the 42 or 840 Billion token 300 \
        dimension common crawl glove vectors. The values can be only 42 or \
        840 and default is 42.
        :param skip_conf: Whether to skip the permission step for downloading \
        the word vectors as it requires user input. True to skip permission.
        :type version: int. Default 42.
        :type skip_conf: bool. Default False
        '''
        if version not in [42, 840]:
            raise ValueError('Common Crawl only come in two version the 840 '
                             'or 42 Billion tokens. Require to choose between '
                             f'42 and 840 and not {version}')

        if name is None:
            name = 'glove 300d {}b common crawl'.format(version)
        vector_file = self.download(skip_conf, version)
        print(f'Loading {name} from file')
        glove_key_vectors = KeyedVectors.load_word2vec_format(vector_file,
                                                              binary=True)
        super().__init__(glove_key_vectors, name=name, unit_length=unit_length,
                         padding_value=padding_value,
                         filter_words=filter_words)

    def _unknown_vector(self):
        '''
        This is to be Overridden by sub classes if they want to return a custom
        unknown vector.

        :returns: A vector for all unknown words. In this case it is a zero
        vector.
        :rtype: numpy.ndarray
        '''

        return np.zeros(self.vector_size, dtype=np.float32)


class GloveWikiGiga(WordVectors):

    def download(self, skip_conf):
        '''
        This method checks if the
        `Glove Wikipedia Gigaword word vectors
        <https://nlp.stanford.edu/projects/glove/>`_
        are already in the repoistory if not it downloads and unzips the word
        vectors if permission is granted and converts them into a gensim
        KeyedVectors binary representation.

        :param skip_conf: Whether to skip the permission step as it requires \
        user input. True to skip permission.
        :type skip_conf: bool
        :returns: A dict containing word vector dimension as keys and the \
        absolute path to the vector file.
        :rtype: dict
        '''

        glove_folder = BELLA_VEC_DIR.joinpath(f'glove_wiki_giga')
        glove_folder.mkdir(parents=True, exist_ok=True)
        current_glove_files = set([glove_file.name for glove_file
                                   in glove_folder.iterdir()])
        all_glove_files = set(['glove.6B.50d.binary',
                               'glove.6B.100d.binary',
                               'glove.6B.200d.binary',
                               'glove.6B.300d.binary'])
        interset = all_glove_files.intersection(current_glove_files)
        # If the files in the folder aren't all the glove files that would be
        # downloaded re-download the zip and unzip the files.
        if interset != all_glove_files:
            can_download = 'yes'
            if not skip_conf:
                download_msg = 'We are going to download the glove vectors '\
                               'this is a large download of ~900MB and takes '\
                               '~2.1GB of disk space after being unzipped. '\
                               'Would you like to continue? If so type '\
                               '`yes`\n>> '
                can_download = input(download_msg)

            if can_download.strip().lower() == 'yes':
                download_link = 'http://nlp.stanford.edu/data/glove.6B.zip'
                glove_zip_path = glove_folder.joinpath('glove_zip.zip')
                # Reference:
                # http://docs.python-requests.org/en/master/user/quickstart/#raw-response-content
                with glove_zip_path.open('wb') as glove_zip_file:
                    request = requests.get(download_link, stream=True)
                    total_size = int(request.headers.get('content-length', 0))
                    print('Downloading Glove Wikipedia Gigaword vectors')
                    for chunk in tqdm(request.iter_content(chunk_size=128),
                                      total=math.ceil(total_size//128)):
                        glove_zip_file.write(chunk)
                print('Unzipping word vector download')
                glove_zip_path = str(glove_zip_path.resolve())
                with zipfile.ZipFile(glove_zip_path, 'r') as glove_zip_file:
                    glove_zip_file.extractall(path=glove_folder)
            else:
                raise Exception('Glove Twitter vectors not downloaded '
                                'therefore cannot load them')

        def add_full_path(file_name):
            file_path = glove_folder.joinpath(file_name)
            return self.glove_txt_binary(file_path)

        return {50: add_full_path('glove.6B.50d.txt'),
                100: add_full_path('glove.6B.100d.txt'),
                200: add_full_path('glove.6B.200d.txt'),
                300: add_full_path('glove.6B.300d.txt')}

    def __init__(self, dimension, name=None, unit_length=False, skip_conf=False,
                 padding_value=None, filter_words=None):
        '''
        :param dimension: Dimension size of the word vectors you would like to \
        use. Choice: 50, 100, 200, 300
        :param skip_conf: Whether to skip the permission step for downloading \
        the word vectors as it requires user input. True to skip permission.
        :type dimension: int
        :type skip_conf: bool. Default False
        '''

        dimension_file = self.download(skip_conf)
        if not isinstance(dimension, int):
            raise TypeError('Type of dimension has to be int not {}'\
                            .format(type(dimension)))
        if dimension not in dimension_file:
            raise ValueError('Dimension avliable are the following {}'\
                             .format(list(dimension_file.keys())))
        if name is None:
            name = 'glove wiki giga {}d'.format(dimension)
        vector_file = dimension_file[dimension]
        print(f'Loading {name} from file')
        glove_key_vectors = KeyedVectors.load_word2vec_format(vector_file,
                                                              binary=True)
        super().__init__(glove_key_vectors, name=name, unit_length=unit_length,
                         padding_value=padding_value,
                         filter_words=filter_words)

    def _unknown_vector(self):
        '''
        This is to be Overridden by sub classes if they want to return a custom
        unknown vector.

        :returns: A vector for all unknown words. In this case it is a zero
        vector.
        :rtype: numpy.ndarray
        '''

        return np.zeros(self.vector_size, dtype=np.float32)



from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='bella_tdsa',
      version='0.3.1',
      description='Target Dependent Sentiment Analysis (TDSA) framework.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/apmoore1/Bella',
      author='Andrew Moore',
      author_email='andrew.p.moore94@gmail.com',
      license='MIT',
      install_requires=[
          'filelock>=3.0.10',
          'Keras>=2.1.3',
          'tensorflow>=1.3.0',
          'scikit-learn==0.19.1',
          'gensim>=3.0.1',
          'requests>=2.18.4',
          'tqdm>=4.23.4',
          'twokenize>=1.0.0',
          'nltk>=3.2.5',
          'stanfordcorenlp==3.7.0.2',
          'ftfy>=5.2.0',
          'ruamel.yaml>=0.15.34',
          'pandas>=0.21.0',
          'networkx>=2.0',
          'tweebo-parser-python-api>=1.0.4',
          'seaborn>=0.8.1',
          'graphviz>=0.8.4',
          'pydot>=1.2.4'
      ],
      python_requires='>=3.6.1',
      packages=find_packages(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.6',
          'Topic :: Text Processing',
          'Topic :: Text Processing :: Linguistic',
      ])



import numpy as np
from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

from bella import neural_pooling

class JoinContextVectors(BaseEstimator, TransformerMixin):

    def __init__(self, pool_func=neural_pooling.matrix_median):
        self.pool_func = pool_func

    def fit(self, context_pool_vectors, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, context_pool_vectors, y=None):
        '''see self.transform'''

        return self.transform(context_pool_vectors)

    def transform(self, context_pool_vectors):
        '''
        Given a list of train data which contain a list of numpy.ndarray one for
        each context. Return a list of train data of numpy.ndarray which are the
        contexts joined together using one of the pool functions.
        '''
        vec_size = context_pool_vectors[0].shape[1]
        train_vectors = []
        len_train_data = len(context_pool_vectors)
        for context_pool_vector in context_pool_vectors:
            train_vectors.append(self.pool_func(context_pool_vector))
        train_vectors = np.asarray(train_vectors).reshape(len_train_data, vec_size)
        return train_vectors
